% Layered network learns by online error-backpropagation

% Neurons in final layer are linear.
% Code uses cell arrays to define chains of different-sized matrices,
% W{} & a{}.

clear all; clc

% Initialize
nx = 2; nl = 3; nn = [nx + 1; 10; 1];  % # inputs, # layers including input, # neurons in layers
ns = 5e4;  % # examples
nw = 0; for l = 2:nl, nw = nw + nn(l - 1)*nn(l); end, nw  % # adjustable weights 
Q = diag(ones(nx, 1));  % used to compute desired output, y = x'*Q*x
y_sum = 0; y_sqr_sum = 0;
W = cell(nl, 1); DW = cell(nl, 1); eta = zeros(nl, 1);
a = cell(nl, 1); L_v = cell(nl, 1);
for layer = 2:nl
  m = 3.465/sqrt(nn(layer - 1));
  W{layer} = m*(rand(nn(layer), nn(layer - 1)) - 0.5);  % set var(w) = 1/#syn so var(v) = 1/4
  DW{layer} = zeros(nn(layer), nn(layer - 1));
  eta(layer) = 0.5/nn(layer - 1);  % learning-rate constant
end
mo = 0.9;  % momentum for all W's except W{nl}
gt_step = ceil(ns/1000);  % graphical step, ensures <= 1000 pts per plot
n = 0;  % index of graph points
DATA = zeros(2, nw + floor(ns/gt_step));  % allocate more than enough memory

% Train
for s = 1:ns
    
  % Set input and desired output
  x = 1.732*(rand(nx, 1) - 0.5);  % choose x with SD 0.5
  y = x'*Q*x;
  y_sum = y_sum + y; y_sqr_sum = y_sqr_sum + y'*y;  % statistics of y
  
  % Compute network's output and error
  a{1} = [x; 0.5];  % bias term to fit y()'s where y(0) <> 0
  for layer = 2:nl - 1, a{layer} = tanh(W{layer}*a{layer - 1}); end  % i.e. a = tanh(v) = tanh(W*a-)
  a{nl} = W{nl}*a{nl - 1};  % neurons in final layer are linear
  y_err = a{nl} - y;  % error
   
  % Adjust weights
  L_v{nl} = y_err';
  W{nl} = W{nl} - eta(nl) * L_v{nl}' * a{nl - 1}';  % i.e. dL/dv * dv/dW
  for layer = nl - 1:-1:2
    L_v{layer} = L_v{layer + 1} * W{layer + 1} .* (1 - a{layer}.*a{layer})';  % i.e. dL/dv+ * dv+/da * Da/Dv 
    DW{layer} = mo*DW{layer} - eta(layer) * L_v{layer}' * a{layer - 1}';  % i.e. dL/dv * dv/dW
    W{layer} = W{layer} + DW{layer};
  end  
  
  % Record variables for plotting
  if (s <= nw) || (mod(s, gt_step) == 0), n = n + 1;
    DATA(:, n) = [s; y_err'*y_err];
  end

end  % for s
DATA = DATA(:, 1:n);

% Compute statistics of y
y_avg = y_sum/ns; y_var = y_sqr_sum/ns - y_avg'*y_avg;
DATA(2, :) = DATA(2, :)/y_var;  % normalized squared error, NSE

% Plot
figure(1), set(gcf, 'Name', 'Backpropagation', 'NumberTitle', 'off');
%DATA(2, :) = sgolayfilt(DATA(2, :), 1, 51);  % smooth data
subplot(2, 1, 1), plot(DATA(1, :), DATA(2, :), 'r'), ylabel('NSE')
line([1 gt_step*n], [0.01, 0.01], 'LineStyle', '--', 'Color', 'k')
ylim([0 0.1])
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
subplot(2, 1, 2), loglog(DATA(1, :), DATA(2, :), 'r'), ylabel('NSE')
xlabel('Example')
line([1 gt_step*n], [0.01, 0.01], 'LineStyle', '--', 'Color', 'k')
ylim([1e-10 1e2])
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
