% State-feedback controller for saccade-like movements of a 1st-order
% plant

clear all; clc

% Set parameters
dur = 4;  % duration of run in s
Dt = 0.01;  % Euler-integration time step in s
wait = ceil(0.5/Dt);  % interval betw target jumps in Euler time steps
a = -0.005/Dt; G = 0.02;  % in state dynamics, q_vel = f + G*u where f = a*q
a_est = 1*a; G_est = 1*G;  % controller's estimates of a & G
DATA = zeros(4, floor(dur/Dt));  % allocate memory

% Initialize
q_star = 0;  % initial desired eye position in radians
q = q_star;  % eye position at time t = 0
u = -a*q/G;  % neural command holds initial eye position
q_vel = 0;  % eye velocity
c = -10;  % coeff of desired dynamics q_vel = c*(q - q_star)
step = 0;  % # time steps since simulation began
n = 0;  % index of graph pts

% Run time loop
for t = 0:Dt:dur, step = step + 1;
    
  % Set target
  if mod(step, wait) == 1, q_star = 2*(rand - 0.5); end
  
  % Compute command
  e = q - q_star;  % eye-position error
  q_vel_star = c*e;  % Hurwitz desired dynamics
  f_est = a_est*q;
  u = G_est\(q_vel_star - f_est);
  
  % Update plant
  f = a*q; q_vel = f + G*u; q = q + Dt*q_vel;
   
  % Record data for plotting
  n = n + 1; DATA(:, n) = [t; q_star; q; u];

end  % for t
DATA = DATA(:, 1:n);

% Plot
figure(1), set(gcf, 'Name', 'Feedback Linearization', 'NumberTitle', 'off');
subplot(2, 1, 1), plot(DATA(1, :), DATA(2, :), 'b:'), ylabel('q'), hold on
subplot(2, 1, 1), plot(DATA(1, :), DATA(3, :), 'b')
ylim([-1.25 1.25])
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
subplot(2, 1, 2), plot(DATA(1, :), DATA(4, :), 'k'), ylabel('u'), xlabel('Time')
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
