% HG_TO_1DOF: Hamiltonian Gradient Time Optimal 1 Degree Of Freedom.

% Learns to move a 1-DOF point-mass roughly time-optimally (i.e. quickly)
% to 0. Obeys the constraint that ||u|| <= 1. Uses damped NLMS to adjust
% J_est --- an estimate of the optimal cost-to-go J* --- down the gradient
% dH_est^2/dw, where H_est is an estimate of the Hamiltonian H*. Uses
% nonlinear gradient descent to adjust u() --- its current policy, which
% is also an estimate of the optimal policy u*() --- down the gradient
% dH_est/dW. 

% Calls sys_simple.m.

% Waits for a button press after drawing first simulation.

clear all; clc

global nq

% Initialize
nq = 1; nx = 2; nu = nq; nz = 10; ns = 1e7; 
dur = 30; Dt = 0.01;  % maximum duration of movt in s; Euler time step in s
hpr = -0.2;  % Hurwitz poly root for initial dynamics
for i = 1:nz, x = 2*(rand(nx, 1) - 0.5); WX(i, :) = x'/(x'*x); end    
w = zeros(1, nz);  % adjustable synapses in model
Q = 3;  % coefficient in cost-rate function
eta_W = 1e-4;
x_test = [-0.5; 0];  % initial state for test movt, used to assess policies but not for learning
sys_ = @sys_simple;  % computes state dynamics f & G
r_ = @(x, u) tanh(x'*Q*x);  % cost rate encourages quick movts
ufl_ = @(x, f, G) G\(2*hpr*x(2) - hpr*hpr*x(1) - f);  % fdbk-lin policy: q_acc = -2*c - c*c*q
phi_ = @(x) tanh(WX*x);  % features, used for both J*-model and policy

% Parametrize the initial policy, i.e. approximate the fdbk-lin policy with
% one of the form u(x, W)
ns1 = 3e4; X = 2*rand(nx, ns1) - 1; Z = zeros(nz, ns1); Y = zeros(nu, ns1);
for s = 1:ns1
  x = X(:, s); Z(:, s) = phi_(x); [f, G] = sys_(x);
  u = ufl_(x, f, G); if max(u) > 1, u, end  % we want ||u|| <= 1
  Y(:, s) = atanh(u);
end
W = (Y*Z')*pinv(Z*Z');  % find best approximation by batch least squares
u_ = @(x, W) tanh(W*phi_(x));  % tanh ensures ||u|| <= 1

% Compute cost of initial policy on test movt
Cost = 0; x = x_test; DATA = zeros(5, 1 + ceil(dur/Dt)); n = 0;
for t = 0:Dt:dur
  u = u_(x, W); r = r_(x, u); Cost = Cost + r*Dt;
  [f, G] = sys_(x); x_vel = [x(nq + 1:end); f + G*u]; x = x + x_vel*Dt;
  n = n + 1; DATA(:, n) = [t; x; Cost; u];
  if r < 1e-3, break, end
end
DATA = DATA(:, 1:n); Cost
figure(1)
subplot(3, 1, 1), plot(DATA(1, :), DATA(2:3, :)), ylabel('q')
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
subplot(3, 1, 2), plot(DATA(1, :), DATA(4, :), 'r'), ylabel('J')
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
subplot(3, 1, 3), plot(DATA(1, :), DATA(5, :)), ylabel('u'), xlabel('Time (ms)')
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
drawnow
waitforbuttonpress;  

% Improve policy
for s = 1:ns
    
  % Adjust J*-model & policy
  x = 2*(rand(nx, 1) - 0.5); u = u_(x, W); [f, G] = sys_(x);
  x_vel = [x(nq + 1:end); f + G*u];
  phi = phi_(x); phi_vel = phi.*(WX*x_vel); H_est = w*phi_vel + r_(x, u);
  w = w - 0.5*H_est*phi_vel'/(phi_vel'*phi_vel);  % adjust J_est (estimate of J*) by damped NLMS
  W = W - eta_W*((1 - u.*u).*((w.*phi')*WX(:, nq + 1:nx)*G)')*phi';  % adjust policy by gradient descent down dH/dWU
    
  % Compute cost of current policy on test movt
  if mod(s, 1e3) == 1
    Cost = 0; x = x_test; DATA = zeros(5, 1 + ceil(dur/Dt)); n = 0;
    for t = 0:Dt:dur
      [f, G] = sys_(x); u = u_(x, W); r = r_(x, u); Cost = Cost + r*Dt;
      x_vel = [x(nq + 1:end); f + G*u]; x = x + x_vel*Dt;
      n = n + 1; DATA(:, n) = [t; x; Cost; u];
      if r < 1e-3, break, end
    end
    DATA = DATA(:, 1:n); Cost
    subplot(3, 1, 1), plot(DATA(1, :), DATA(2:3, :)), ylabel('x')
    set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
    subplot(3, 1, 2), plot(DATA(1, :), DATA(4, :), 'r'), ylabel('J')
    set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
    subplot(3, 1, 3), plot(DATA(1, :), DATA(5, :)), ylabel('u'), xlabel('Time')
    set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
    drawnow
  end  
  
end  % for s
