% RLS with Gaussian features

% Uses a "centers matrix" CEN to position the Gaussian features:
% the i'th Gaussian is centered on the i'th column vector of CEN, i.e.
% phi[i](x) = exp(-||CEN(:, i) - x||^2 / 2*variance of Gaussian).
% We define D = CEN - x*o & nidvarg = -1/(2*variance of Gaussian), so
% phi(x) = exp( nidvarg * sum(D.*D)' ).

% Doesn't sparsify.

% Works with wide Gaussians, i.e. small nidvargs.

clear all; clc

% Initialize
nx = 2; nf = 10; ns = 1e2*nf;  % # elements in input x; # features; # examples
dom = 1;  % radius of domain
Q = eye(nx);  %2*diag(mod(1:nx, 2) - 0.5);  % used to compute y = x'*Q*x
y_sum = 0; y_sqr_sum = 0;
nidvarg = -0.01/dom*dom;  % neg inv of double the variance of the Gaussians
CEN = zeros(nx, nf);  % allocate memory for centers matrix
w = zeros(1, nf);  % initial weights
P = 1e10*eye(nf);  % soft initialization
gt_step = ceil(ns/1000);  % graphical step, ensures <= 1000 pts per plot
DATA = zeros(2, nf + floor(ns/gt_step)); 
n = 0;  % index of graph pts

% Show complexity
nw_adj = nf  % # adjustable weights
nn = nf  % # neurons excluding x & y_est
nsyn = nf*(1 + nx + nf)  % adjustable weights, centers, and P
%serial_learn_time = nf^3

% Choose feature centers
for i = 1:nf, x = 2*dom*(rand(nx, 1) - 0.5); CEN(:, i) = x; end
o = ones(1, nf);

% Train
for s = 1:ns
  x = 2*dom*(rand(nx, 1) - 0.5); y = x'*Q*x;  % input and desired output
  y_sum = y_sum + y; y_sqr_sum = y_sqr_sum + y*y;  % statistics of y
  D = CEN - x*o; phi = exp(nidvarg*sum(D.*D)');  % feature vector
  y_est = w*phi; y_err = y_est - y;  % estimate of y; model error
  Pphi = P*phi; k = Pphi'/(1 + phi'*Pphi); P = P - Pphi*k; w = w - y_err*k;  % RLS
  if (s <= 10) || (mod(s, gt_step) == 0), n = n + 1;
    DATA(:, n) = [s; y_err'*y_err];  % record data for plotting
  end  
end  % for s
DATA = DATA(:, 1:n);

% Compute statistics of y
y_avg = y_sum/ns; y_var = y_sqr_sum/ns - y_avg*y_avg;
DATA(2, :) = DATA(2, :)/y_var;  % normalize squared model errors  

% Plot
figure(1), set(gcf, 'Name', 'RLS Gaussian', 'NumberTitle', 'off');
subplot(2, 1, 1), plot(DATA(1, :), DATA(2, :), 'r'), ylabel('NSE')
xlim([0 gt_step*n]), ylim([0 0.1])
line([0 gt_step*n], [0.01 0.01], 'LineStyle', '--', 'Color', 'k')
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.005])
subplot(2, 1, 2), loglog(DATA(1, :), DATA(2, :), 'r'), ylabel('NSE'), xlabel('Example')
xl = xlim; ylim([1e-8 1e2]); 
line([xl(1), xl(2)], [0.01, 0.01], 'LineStyle', '--', 'Color', 'k')
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.005])
